/*
  Enviar, Recibir, usa CRC-32
*/
#define DEBUG 1	//DEBUG solo es para probar el CRC, no para recibir/transmitir

#include <dos.h>
#include <time.h>
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>

#define IRQ 0x60	//numero de la interrupcion que usamos normalmente

char far *bloque;	//apuntador al bloque recibido
int len;		//longitud del bloque recibido
int bandera=0;		//para saber si ya termin de llegar
int handle;		//para release_type
static long int numero;	//residuo del CRC
int Derr;		//para send_pkt, bandera de que hubo error

const int CRCLEN=4;	//cuantos bytes tendra el CRC
//polinomio=0000 0100 1100 0001 0001 1101 1011 0111
//26 23 22 16 12 11 10 8 7 5 4 2 1 0, se asume que el bit 32 tambien vale 1
#if !DEBUG
const long int POLINOMIO=0X04C11DB7;
#else
const long int POLINOMIO=0x09;	//polinomio de prueba (cualquiera)
#endif

union b{		//para el CRC
  char arr[CRCLEN];
  long int CRC;
} *a;

void accestype(void *);			//Recibir
void interrupt receiver();		//Recibir
void release_type(int);			//Recibir
void swap_dir(char far *buff);		//Recibir
long int crc(char z, char f=1);		//CRC
int send_pkt(char far *, unsigned);	//Enviar

void main(void)
{
	int i, j;
	char temporal[CRCLEN];

	clrscr();
#if !DEBUG
	//Recibir
	accestype(receiver);	//se prepara para recibir
	printf("Esperando recibir un bloque...");
	while(bandera==0);	//hasta que termina de recibir (ver receiver)
	release_type(handle);	//termina la recepcion
	printf("ha llegado un paquete! :)\n\a");
	for(i=0 ; i<len ; i++)	//imprime el bloque recibido en hexadecimal
		printf("%X ", (unsigned char)bloque[i]);
	swap_dir(bloque);	//intercambia las direcciones para asi poder mandarlas a
	printf("\n\nPresiona una tecla para continuar...");//la maquina de donde vienen
	getch();
#else
	srand(time(NULL));
	len=5;
	bloque=new char[len+CRCLEN];
	for(i=0; i<len; i++)
		bloque[i]=rand()%256;
	printf("Paquete:\n");
	for(i=0 ; i<len ; i++)	//imprime el bloque en hexadecimal
		printf("%X ", (unsigned char)bloque[i]);
#endif
	//Calcular CRC
	crc(0,0);			//inicializa el residuo a 0
	for(i=0; i<len; i++)
		crc(bloque[i]);		//calcula el crc para cada byte del bloque
	a=(union b*)&bloque[len];	//apunta a ultimo elemento del bloque
	for(i=0; i<CRCLEN; i++)
		a->CRC=crc(0);		//a->CRC=crc de '\0'+crc ya calculado
	for(j=CRCLEN-1,i=len; i<len+CRCLEN; i++,j--)
		temporal[j]=bloque[i];	//voltea el CRC pq queda al reves
	for(i=0; i<CRCLEN; i++)
		bloque[len+i]=temporal[i];//pega el CRC ya volteado
	printf("\n\nPaquete a reenviar:\n");
	for(i=0; i<len+CRCLEN; i++)//imprime el bloque con el CRC en hexadecimal
		printf("%X ", (unsigned char)bloque[i]);
	printf("\n\nTamao del paquete: %d bytes (+4 del CRC)",len);
#if DEBUG
	//Esta parte solo verifica que el crc este bien calculado
	crc(0,0);
	for(i=0; i<len+CRCLEN; i++)
		crc(bloque[i]);
	printf("\nNumero (debe valer cero): %ld\n",numero);
#else
	//Reenviar
	printf("\nListo para reenviar el paquete...");
	getch();
	if(send_pkt(bloque,len+CRCLEN))
		printf("\n\nError al transmitir! >:(\n\a");
	printf("\n\nListo! =O)\n\a");
#endif
	getch();
	delete bloque;
	exit(0);
}

/*RECIBIR*/
/*Prepara para recibir*/
void accestype(void *receiver)
{
	union REGS inregs, outregs;
	struct SREGS segregs;
//	char far *type1;

	inregs.h.ah=0x02;		//id de accesstype
	inregs.h.al=1;			//if_class
	inregs.x.bx=0xFFFF;		//if_type
	inregs.h.dl=0;			//if_number
	inregs.x.cx=0;			//typelen
	segregs.es=FP_SEG(receiver);	//far *receiver
	inregs.x.di=FP_OFF(receiver);	//far *receiver
//	inregs.x.si = FP_OFF(*type1);
//	segregs.ds = FP_SEG(*type1);
	int86x(IRQ, &inregs, &outregs, &segregs);

	handle=outregs.x.ax;		//handle
}

/*Recibe*/
void interrupt receiver(void)
{
	int far *pila;

	pila=(int*)MK_FP(_SS,_BP);	//"declara" una pila
	len=*(pila+6);			//longitud del bloque recibido
	if(*(pila+8)==0)		//mientras no termina...
	{
		bloque=new char[len+CRCLEN];
		*(pila+4)=FP_SEG(bloque);	//direccion del bloque
		*(pila+1)=FP_OFF(bloque);
	}
	else if(*(pila+8)==1)
		bandera=1;		//termino de recibir
}

/*Termina de recibir*/
void release_type(int handle)
{
	union REGS inregs, outregs;
	struct SREGS segregs;
//	char far *type1;

	inregs.h.ah=0x03;		//id de release_type
	inregs.x.bx=handle;		//handle
	int86x(IRQ, &inregs, &outregs, &segregs);
}

/*Intercambia las direcciones en la trama, destino/origen*/
void swap_dir(char *buff)
{//esto se hace para que ahora podamos transmitir hacia la misma terminal que
	char hold[6]={0x00};				//nos envio el bloque

	for(int i=0; i<6; i++)
		hold[i]=buff[i];
	for(i=0; i<6; i++)
		buff[i]=buff[i+6];
	for(i=0; i<6; i++)
		buff[i+6]=hold[i];
}

/*CRC-32*/
/*Calcula el residuo con CRC-32 de 'z'*/
long int crc(char z, char f)	//por el prototipo, f vale 1 por defecto
{
	const char bmschar=0x80;
	const long int bmsnum=0x80000000;
	char aux, i;
	long int j;

	if(f==0)	//para inicializar el residuo
	{
		numero=0;
		return numero;
	}
	for(int x=0;x<8;x++)	//una vez por cada bit del caracter z
	{
		i=z&bmschar;	//si el bit mas significativo de z vale 1
		j=numero&bmsnum;//si el bit mas significativo de numero vale 1
		if(i)
			aux=1;
		else
			aux=0;
		numero<<=1;	//inyecta 1 cero a la derecha (menos significativo)
		numero+=aux;	//suma 1 al residuo si el BMS de z vale 1
		z<<=1;		//inyecta 1 cero a la derecha de z
		if(j)		//si el BMS de numero valia 1...
			numero=numero^POLINOMIO;//..hace la suma con XOR
	}
	return numero;	//residuo calculado de z
}

/*ENVIAR*/
/*Send packet*/
int send_pkt(char far *buffer, unsigned length)
{
	union REGS regs;
	struct SREGS sregs;

	segread(&sregs);
	regs.h.ah=0x04;			//id de send_pkt
	sregs.ds=FP_SEG(buffer);	//direccion del buffer
	sregs.es=FP_SEG(buffer);
	regs.x.si=FP_OFF(buffer);
	regs.x.cx=length;		//tamao del buffer a enviar

	int86x(IRQ,&regs,&regs,&sregs);

	if(regs.x.cflag)		//si hay error...
	{
		Derr=regs.h.dh;
		return -1;
	}
	return 0;
}
